﻿# Eijiro117.csv2html-v4.pl を改造．　　　　'12.7.5
# 発音記号の表示と句動詞（群動詞）の検索に対応．
# 単語の型変化処理で不要なゴミ（ピリオドや過去・過分項目）が残る問題を修正
# 引数をCSVファイル名と省略可能なオプションのみに変更．
#
# csv2html.pl eijiro.csv         通常処理（発音記号は解説先頭に表示）
# csv2html.pl eijiro.csv -ps     発音記号（Phonetic Symbols）を非表示にする．
# csv2html.pl eijiro.csv +ps     発音記号を解説末尾に表示する．

#!/usr/bin/perl
#
use strict;
use warnings;
use utf8;
use Encode;
use FileHandle;
use Text::CSV_XS;
use PerlIO::encoding;
use Encode qw(:fallbacks);
use HTML::Entities;

sub getInflections(**)
{
  my ($arg, $infs) = @_;
  my $cnt;

  @{$infs} = ${$arg} =~ /【変化】(.*?)(【|◆|<br>|\r).*$/;

  if ($#{$infs} >= 0) {

    $infs->[0] =~ s/（まれ）//g;
    $infs->[0] =~ s/または/\|/g;

    if ($infs->[0] =~ /[\[\];,] | (\s-\s)/) {           # [ ]で囲まれた項目と ; , -
      $infs->[0] =~ s/\[/\| /g;                         # [ ]を外し境界に | を挿入
      $infs->[0] =~ s/\]//g;
      $infs->[0] =~ s/;/\|/g;                           # ; \s-\s , を | に置換
      $infs->[0] =~ s/\s-\s/ \| /g;
      $infs->[0] =~ s/,/ \|/g;
    }

    if ($infs->[0] =~ /[\(\)]/) {                       # ( )で囲まれた項目
      $infs->[0] =~ s/ \| / /g;                         #  | を一度スペースに変換
      $infs->[0] =~ s/(\w+?\(.\)\w+?[\s、])/$1$1/g;     # ( )で囲まれた項目をコピー
      $cnt = 0;

      foreach my $ch (split //, $infs->[0]) {           # ( )で囲まれた項目を数える
        if ($ch eq '(') {
          $cnt++;
        }
      }
      $cnt /= 2;

      for (my $i=0; $i<$cnt; $i++) {                    # ( )の部分を先頭から順に変換＆削除
        $infs->[0] =~ s/\(.\)//;
        $infs->[0] =~ s/\(//;
        $infs->[0] =~ s/\)//;
      }
      $infs->[0] =~ s/ / \| /g;                         # スペースをすべて' | 'に変換
    }

    if ($infs->[0] =~ /《形》/) {                       # 比較級の項目を削除  '12.7.9
      if ($infs->[0] =~ /more/) {
        $infs->[0] =~ s/《形》.+?、/《形》、/;
      }
    }

    $infs->[0] =~ s/《.+?》//g;
    $infs->[0] =~ s/\.//g;                              # 不要なピリオドを削除
    $infs->[0] =~ s/過去・過分＝[a-zA-Z\-]+//g;         # 不要な過去・過分項目を削除
    $infs->[0] =~ s/ZXZX.+XZXZ//g;                      # 不要なリンク項目を削除
    $infs->[0] =~ s/、/ \| /g;
    # $infs->[0] =~ s/ +(\| +)*?《.+?》/ \| /g;
    @{$infs} = grep(/\w/, split(' \| ',$infs->[0]));    # 熟語の複数形に対応  '12.7.9
    # print STDERR encode('shift_jis', "inflections: " . join('//', @{$infs}) . "\n");
    # print encode('shift_jis', "inflections: " . $infs->[0] . "\n");
    # print encode('shift_jis', "rest: " . $infs->[1] . "\n=====\n");
  }
}

sub Substitute(*)
{
  my $arg = $_[0];

  ${$arg} =~ s/【＠】.*?【/【/g;
  # ${$arg} =~ s/【変化】《.+?》.*?(、|<br>)//g;
  ${$arg} =~ s/【変化】.*?(、|<br>)//g;
  ${$arg} =~ s/【大学入試】//g;
  ${$arg} =~ s/ \| /\//g;
# ${$arg} =~ s/【(.+?)】/[__START_BOLD__$1__END_BOLD__]/g;
# ${$arg} =~ s/《(.+?)》/[__START_BOLD__$1__END_BOLD__]/g;
  ${$arg} =~ tr/。、【】《》［］％＿（）＝＾/.,[]<>()%_()=^/;
  ${$arg} =~ s/◆ */ /g;
  ${$arg} =~ s/-([0-9]+)\]/$1\]/g;

  ${$arg} =~ s/<br>/ /g;
  ${$arg} =~ s/\r/ /g;

}

sub printElements(*)
{
  my($array) = @_;
  my $i = 0;
  my $str = "";

# /* $#{$array} */
for ($i = 0; $i <= 1 ; $i++) {
    $str = $array->[$i];
    Substitute(\$str);
    print encode('shift_jis', "[$i] $str\n");
  }
}

sub getStringOfLevels(*)
{
  my $array = $_[0];
  my $retStr = "";
  
  for (my $x = 0; $x <= $#{$array}; $x++) {
    $retStr .= ("level[$x]:" . $array->[$x] . "\n");
  }
  return $retStr;
}

my %linkedHash;

sub replaceLink($)
{
  my $idx = $_[0];
  my $id = $linkedHash{$idx};
  if (defined $id) {
    # print STDERR encode('shift_jis', "** $idx, $id\n");
    return "ZXZXZXZ${id}ZXZXZXZ${idx}ZXZXZXZ\n";
  } else {
    return "[参照]$idx\n";
  }
}

sub restoreLink($$)
{
  my ($id, $idx) = @_;
  
  return "<a href=\"#${id}\">[参照]${idx}</a>\n";
}

  


#  =======
#  M A I N
#  =======
#
die "please specify the name of file for input.\n" if ($#ARGV < 0);
#die "please specify the name of your Dictionary.\n" if ($#ARGV < 1);

my $charCodeOfCSV = 'UTF-16LE';
#my $charCodeOfCSV = 'cp932';
#$charCodeOfCSV = $ARGV[2] if $#ARGV >= 2;

my $unsafe_chars = '<>&"';

my $fileName = $ARGV[0];
my $comOption = '';
if (@ARGV > 1) {
  $comOption = $ARGV[1];
}
my $dictionaryName = decode('shift_jis', encode_entities('Eijiro-WIK:4-ITAI', $unsafe_chars));
#my $dictionaryName = decode('shift_jis', encode_entities($ARGV[1], $unsafe_chars));

my ($wordsInKey) = ($dictionaryName =~ /WIK:(\d+)/);
my ($idiomsToAnotherIndex) = ($dictionaryName =~ /ITAI/ ? 1 : undef);
$wordsInKey = 1 if !defined($wordsInKey);

my $fileOfLinkedWords = "linked.txt";
my $hasLinkedWords = 0;
my $linkedId = "LK0000000";
if (-f $fileOfLinkedWords) {
  $hasLinkedWords = 1;
  open my $fhLink, "<:encoding(UTF-8)", $fileOfLinkedWords || ($hasLinkedWords = 0);
  if ($hasLinkedWords) {
  print STDERR encode('shift_jis', "$fileOfLinkedWords has been opened!!\n");
    while (<$fhLink>) {
      chomp;
      # print STDERR encode('shift_jis', "$_: $linkedId\n");      
      $linkedHash{$_} = $linkedId;
      $linkedId++;
    }
    close $fhLink;
  }
}

my $htmlFirstPage = <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">
<?xml version="1.0"?><HTML xmlns:idx="s" xmlns:mbp ="s" xmlns:xlink ="http://www.w3.org/1999/xlink">
<head>
  <title>$dictionaryName</title>
  <meta http-equiv=Content-Type content="text/html; charset=utf8">
</head>
<body topmargin="0" bottommargin="0" rightmargin="2" leftmargin="2">
  <div bgcolor="#0835C4" border="1">
    <h2 align="center"><span color="white">$dictionaryName</span></h2>
  </div>
  <div bgcolor="#FA8072" border="1" align="center">
    <idx:orth>&nbsp;TOP PAGE&nbsp;</idx:orth>
  </div>
  <p align="center"><font color="blue"><br>
    <a onclick="index_search('','検索','','','',24)"><b>検索</b></a> |
    <a onclick="index_search('','検索','*','','',24)"><b>*/?</b></a>
  </font></p>
<mbp:pagebreak>
EOF

print encode('utf-8', $htmlFirstPage);

open my $fh, "<:encoding($charCodeOfCSV)", $fileName || die "error, Cannot open $fileName¥n";
my $csv = Text::CSV_XS->new({ binary => 1, eol => $/ }); #日本語を読み込むときはbinaryを1にする

# 一行ごとに読み込む
my $i = 0;
my $maxLengthOfEntry = 0;
my @levels = (0..10);
my $indexName = "";
foreach my $j (@levels) { $levels[$j] = 0; }

my $item7 = '';                                     # 発音記号
my $keyBuff = '';                                   # 見出単語
my @inflBuff = ();                                  # 単語変化リスト

LOOP_OF_GETLINE:
while (my $row = $csv->getline($fh)) {
  my @fields = @$row;
  if ($comOption ne '-ps') {
    $item7 = $row->[6];                             # CSVから発音記号をコピー
  }

  # last if ($i > 4000);
  next if ($i == 0 && $fields[3] eq 'level');

  if (defined $wordsInKey) {
    my @words = split(' ', $fields[0]);
    next LOOP_OF_GETLINE if ($#words >= $wordsInKey);
    if ($idiomsToAnotherIndex) {
      $indexName = ($#words > 0 ? ' name="idiom" ' : "");
    }
  }
 
  $i++;

  my $targetMark = "";
  if ($hasLinkedWords) {
    my $linkedId = $linkedHash{$fields[0]};
    if (defined $linkedId) {
      $targetMark = "<a name=\"${linkedId}\"></a>\n";
      # $targetMark = "<a name=\"${linkedId}\" />\n";
    }
    $fields[1] =~ s/<→(.*?)>/replaceLink($1)/ge;
  }

  my ($escapedKey, $key, $explain, $entry); 
  $escapedKey = $key = encode_entities($fields[0], $unsafe_chars);
  $escapedKey =~ s/\\/\\\\/g;
  $escapedKey =~ s/'/\\'/g;

  my (@inflections, $inflectionTags);
  $explain = $fields[1] . '<br>' . $fields[2];

  $explain =~ s/\n/<br>/g;

  getInflections(\$explain, \@inflections);
  $inflectionTags = "";
  if ($#inflections >= 0) {
    if ( !($inflections[0] =~ /-/) ) {
      $keyBuff = $key;                      # 見出単語を保存
      @inflBuff = @inflections;             # 単語変化リストを保存
    }
    foreach my $inf (@inflections) {
      $inflectionTags .= "<idx:iform value=\"$inf\"></idx:iform>\n";
    }
    $inflectionTags = "\n<idx:infl>\n$inflectionTags</idx:infl>\n";
  }
  Substitute(\$explain);

# 発音記号の追加と句動詞（群動詞）の変化に対応　'12.2.3

  if ($item7 ne "") {                       # 発音記号あり
    if ($comOption ne '+ps') {
      $explain = $item7 . " " . $explain;
    }
    else {
      $explain = $explain . " [発音] " . $item7;
    }
  }
  if ($explain =~ /(句動|句自動|句他動)/) {     # 句動詞項目
    $inflectionTags = "";

    if ($key =~ /^$keyBuff/) {
      my $tailKey = $key;
      $tailKey = substr($tailKey, length($keyBuff));
      foreach my $inf (@inflBuff) {
        $inflectionTags .= "<idx:iform value=\"$inf" . "$tailKey\"></idx:iform>\n";
      }
      $inflectionTags = "\n<idx:infl>\n$inflectionTags</idx:infl>\n";
    }
  }
  else {                                    # 動詞の変化する熟語に対応　'14.1.17
    if ($key =~ / /) {
      if ($explain =~ /(る,|る\s|す,|す\s|く,|く\s|う,|う\s)/) {
        $inflectionTags = "";

        if ($key =~ /^$keyBuff/) {
          my $tailKey = $key;
          $tailKey = substr($tailKey, length($keyBuff));
          foreach my $inf (@inflBuff) {
            $inflectionTags .= "<idx:iform value=\"$inf" . "$tailKey\"></idx:iform>\n";
          }
          $inflectionTags = "\n<idx:infl>\n$inflectionTags</idx:infl>\n";
        }
      }
    }
  }
  $indexName = "";                          # すべての熟語を見出しに変更

#  my $len = length($explain);
#  $maxLengthOfEntry = $len if ($len > $maxLengthOfEntry);
  print STDERR "[$i]\r" unless ($i % 2000);

  print encode('utf-8', $targetMark);
  encode_entities($explain, $unsafe_chars);
  $explain =~ s/ZXZXZXZ(.*?)ZXZXZXZ(.*?)ZXZXZXZ/restoreLink($1, $2)/ge;

#    $pieceOfExplain =~ s/__START_BOLD__/<b>/g;
#    $pieceOfExplain =~ s/__END_BOLD__/<\/b>/g;
  $entry = <<EOF;
<idx:entry${indexName}>
  <h3>&bull;<idx:orth>$key$inflectionTags</idx:orth></h3>
  $explain
</idx:entry><br />
EOF

  print encode('utf-8', $entry);

}
# print STDERR "[$i:$maxLengthOfEntry] ", getStringOfLevels(\@levels), "\r";
$fh->close();

print <<EOF;
  </body>
</html>
EOF

print STDERR "\n";
