#
#　英辞郎のCSVファイルから、いらない単語＆項目をカットする
#

use utf8;
use Encode;


if (@ARGV < 1) {
    die "Parameter Err";
}

open(IN, "<:encoding(UTF-16LE)", $ARGV[0]) || die "$ARGV[0] : Open Err";
#open(OUT, ">:raw:encoding(UTF-16LE)", $ARGV[1]) || die "$ARGV[1] : Open Err";

binmode STDOUT, ':raw';                 # 標準出力にバイナリモードを設定

@out_buff = ();                         # 書出用バッファ
#@bak_buff = ();                        # バックアップ用バッファ

$pre_word = "";                         # 1項目前の単語
$word_type = 0;                         # 品詞フラグ 1:前・接・副 / 0:それ以外
$word_chk = "";                         # 前置詞or接続詞の単語を一時保存
$main_count = 1;
                                        # 群動詞チェック用バッファ　'14.1.17
@prep_buff = ('about','across','after','away','by','down','for','from','into','off','on','out','up','with');
$main_word = "";                        # 熟語の親となる単語
$main_level = '0';                      # 熟語の親となる単語のLevel

$head_word = '';                        # 見出し語
@infl_list = ();                        # 語形変化バッファ
%infl_hash = ();                        # リンク作成用：語形変化=>原形データ
$idiom_flg = 0;                         # 見出し語の種類  1:熟語 / 0:単語
$link_flg = 0;                          # リンク挿入の有無  1:有り / 0:無し
%gloss_hash = ();                       # 削除用：見出し語=>訳語データ
$word_level = '0';                      # 標準語彙水準(SVL) Lv1～12

sub CntChar                     # 文字列に含まれる文字の数を返す
{
    my $buff = $_[0];           # 文字列
    my $char = $_[1];           # 文字
    my $len = $_[2];            # 文字列の長さ

    $cnt = 0;

    my $wkbuf = substr($buff, 1, $len);

    foreach my $ch (split//, $wkbuf) {
        if ($ch eq $char) {
            $cnt++;
        }
    }
    $cnt;
}

sub PutLine                     # 書出バッファに１行ずつコピーする
{
    my ($outbuf, $inbuf) = @_;        # 書出バッファ、読込バッファ１行分

    my $line = @$outbuf;
    $$outbuf[$line] = $$inbuf;
}

sub ChkIdiom                    # 熟語かどうかチェック
{
    my ($buff) = @_;
    my $flg = 0;

    if ($buff =~ /^"[\w\-',]+\s\w/) {
        $flg = 1;
    }
    return $flg;
}

sub LinkItem            # 原形へのリンクデータ作成
{
    my ($list, $hash, $word) = @_;
    my $link = '';

    foreach my $item (@$list) {
        if ($item eq $word) {
            $link = " <→$$hash{$word}>";
            if ($word eq $$hash{$word}) {
                $link = '';
            }
            return $link;
        }
    }
    return $link;
}

sub SubstLink            # 解説の末尾に原形へのリンクを挿入する
{
    my ($list, $hash, $buff, $word) = @_;
    my $res = 0;

    if ($word =~ /[\w\-]+ed/ || $word =~ /[\w\-]+ing/ || $word =~ /[\w\-]+s/) {
        my $link = LinkItem(\@$list, \%$hash, $word);
        if ($link) {
            $$buff =~ s/","[^"]*",\d+,\d+,\d+,".*"/$link$&/;
            $res = 1;
        }
    }
    return $res;
}

sub InsertLink          # 見出語の解説末尾に原形へのリンクを挿入
{
    my ($buff) = @_;

    if ( !$idiom_flg && !$link_flg) {
        SubstLink(\@infl_list, \%infl_hash, \$$buff, $head_word);
    }
}

sub PrintBuff
{
    my ($buff) = @_;                  # 書出バッファ

    if (@$buff) {
        for ($cnt=0; $cnt<@$buff; $cnt++) {
            InsertLink(\$$buff[$cnt]);
            print encode('UTF-16LE', $$buff[$cnt]);
#            print OUT "$$buff[$cnt]";
        }
    }
}

sub GetLebel                  # バッファ中の語彙レベルを返す
{
    my ($buff) = @_;
    my $level = '0';

    if ($buff =~ /",(\d+),\d+,\d+,/) {
        $level = $1;
    }
    return $level;
}

sub CheckRemoveItem         # 強制削除する項目
{
    my ($hlist, $buff) = @_;
    my $flg = 0;

    if ( !$idiom_flg ) {
        return $flg;
    }

    if ($word_level ne '0') {
        return $flg;
    }

    my ($word, $gloss) = each(%$hlist);
    if ($word eq '' || $gloss eq '') {
        ($word, $gloss) = each(%$hlist);
        return $flg;
    }
    if ($buff =~ /$word/ && $buff =~ /$gloss/) {
        $flg = 1;
    }
    ($word, $gloss) = each(%$hlist);
    return $flg;
}

sub InfoItem
{                        # 書出情報、見出単語、見出文字数、書出バッファ
    my ($info, $ent, $len, $spc, $buff) = @_;

    if ($$buff[0] =~ /<→/ && !$link_flg) {     # リンクのある単語をカット
        $$info[0] = 1;
        if ($$buff[0] =~ /【(形|名|助|他|自|動)/) {    # 形容詞、名詞、動詞は追加
            $$info[1] = 1;
        }
        if ($word_level ne '0') {
            $$info[1] = 1;
        }
        return;
    }

    if ( CheckRemoveItem(\%gloss_hash, $$buff[0]) ) {   # 強制削除する項目
        $$info[0] = 1;
        return;
    }

    if ($spc == 1) {                            # 使用頻度の高い2語の熟語
        if ($$buff[0] =~ /、～/ || $$buff[0] =~ /\w be/) {
            $$info[1] = 1;
            return;
        }
    }

    if ($main_level ne '0') {                   # 2語の群動詞　'14.1.17
        if ($spc == 1) {
            if ($ent =~ /$main_word/) {
                foreach $prep (@prep_buff) {
                    if ($ent =~ / $prep/) {
                        $$info[1] = 1;
                        return;
                    }
                }
            }
        }
    }

    for ($cnt=0; $cnt<@$buff; $cnt++) {
        if ($$buff[$cnt] =~ /【(句動|句自動|句他動)/) {     # 強制追加項目
            $$info[2] = 1;
            last;
        }

        if (3 > $spc && $spc > 0) {                         # 使用頻度の高い熟語
            if ($$buff[$cnt] =~ /^・\w/ || $$buff[$cnt] =~ /【2】/) {
                $$info[2] = 1;
                last;
            }
        }

                # 専門項目を加える（物理、数学、計算機、internet、電気電子etc.）
        if ($$buff[$cnt] =~ /《(物|数|コ|イ|電|生|地|植)/) {
            $$info[3] = 1;
        }

        if ($word_level ne '0') {      # 語彙レベル1以上は追加
            $$info[3] = 1;
        }

        if ($$buff[$cnt] =~ /【(前|接続|副)/) {
            if (!($word_type)) {
                $word_type = 1;
                $word_chk = substr($$buff[0], 0, $len);
                $word_chk =~ s/\(.*//;          # free(3)など単語の()を削除
            }
            $$info[4] = 1;
        }
        if ($$buff[$cnt] =~ /～/) {
            $$info[5] = 1;
        }
        if ($$buff[$cnt] =~ /【(形|名|助|他|自|動)/) {
            $$info[6] = 1;
        }
                                                            # 強制除外項目
        if ($$buff[$cnt] =~ /【(人名|地名|映画|著作|組織|URL)/) {
            if (!($$info[6])) {
                $$info[7] = 1;
            }
        }
    }

}

sub StatItem        # 戻り値  1: 書出項目 / 0: 書出除外項目
{
        # 書出情報、見出先頭１語、見出文字数、単語空数、書出バッファ
    my ($info, $wdbuf, $len, $spc, $buff) = @_;

    my $stat = 0;

    if ($$info[1] | $$info[2]) {        # 強制的に追加する項目
        $stat = 1;
    }
    else {
        if ($$wdbuf eq $pre_word) {             # 同じ単語の熟語は除外
            if ($word_type == 1) {              # 前置詞or接続詞or副詞or語彙Lev≧1
                if ($$info[5]) {                # ～のある項目
                    $stat = 1;
                }
                else {
                    if ($$spc == 1) {           # ２語の熟語
                        $stat = 1;
                    }
                    else {
                        $stat = 0;
                    }
                }
            }
            else {
                if ($$info[3] | $$info[6]) {
                    $stat = 1;
                }
                else {
                    $stat = 0;
                }
            }
        }
        else {                                  # 違う単語と熟語は追加
            $stat = 1;
        }

        if ($$info[0] | $$info[7]) {            # ただし強制除外項目は追加しない
            $stat = 0;
        }

        if (($$len > 25) || ($$spc > 2)) {      # 25文字越え or 4語以上の項目はカット
            $stat = 0;
        }
        if (($$len > 20) && ($$spc > 2)) {      # 20文字を越える3語以上の熟語もカット
            $stat = 0;
        }
    }

    if ($$buff[0] =~ /~|__/) {                  # ~ or __ の入った熟語を強制除外
        $stat = 0;
    }
    $stat;
}

sub ChkItem
{
    my ($buff) = @_;                # 書出バッファ

    my $stat = 0;       # 戻り値  0: 除外項目 ／ 1: 追加項目
    my @info = ();      # 書出情報 [0]～[8]
    my $wdlen;

    if (@$buff) {

        if ($$buff[0] =~ /^"[^a-zA-Z]/) {        # アルファベット以外を除外
            return ($stat = 0);
        }

        my $len = index($$buff[0], '"', 1) - 1;     # 見出し項目（単語＆熟語）
        my $spc = CntChar($$buff[0], ' ', $len);
        my $entbuf = substr($$buff[0], 1, $len);
        
        if ($spc) {
            $wdlen = index($$buff[0], ' ', 1);      # 単語１語
        }
        else {
            $wdlen = $len + 1;
        }
        my $wdbuf = substr($$buff[0], 0, $wdlen);

        if ($word_type) {
            if (!$spc) {                        # 品詞フラグをクリア
                if (!($entbuf =~ /-/)) {
                    $word_type = 0;
                    $word_chk = "";
                }
            }
        }

        InfoItem(\@info, $entbuf, $wdlen, $spc, \@$buff);
        $stat = StatItem(\@info, \$wdbuf, \$len, \$spc, \@$buff);

####
#if ($$buff[0] =~ /^\"as you wish/) {
#if ($main_count > 233400) {
#  print "stat: $stat  [$entbuf]    {$wdbuf} : {$pre_word}\n";
#  print "word_type: $word_type  info:  ($info[0]) ($info[1]) ($info[2]) ($info[3]) ($info[4]) ($info[5]) ($info[6]) ($info[7])\n";
#  print "$main_count:  [$pre_word]  {$word_chk}\n";
#  $dbg = <STDIN>;
#}
####

        $$buff[0] =~ /^"[\w\-]+/;         # 単語の先頭を保存
        $pre_word = $&;

        if ($word_type == 1) {                  # 前／接／副詞の熟語チェック
            if ($pre_word =~ /$word_chk/i) {
                $word_type = 1;
            }
            else {
                $word_type = 0;
                $word_chk = "";
            }
        }
    }
    $stat;
}

sub FormSample
{
    my ($outbuf, $inbuf) = @_;          # 書出バッファ、読込１行分の文字列

    my $len = index($$inbuf, '。');
    if ($len < 0) {
        $len = index($$inbuf, '」');
    }

    if ($len > 160) {                   # 160文字を越える例文を削除
        if ($$inbuf =~ /","/) {
            $len = index($$inbuf, '","');
            my $wkbuf = substr($$inbuf, $len);
            PutLine(\@$outbuf, \$wkbuf);
        }
        else {
            # 書出バッファにコピーしない
        }
    }
    else {
        PutLine(\@$outbuf, \$$inbuf);
    }
}

sub GetLine
{
    my ($outbuf, $inbuf) = @_;              # 書出バッファ、読込１行分の文字列

    if ($$inbuf =~ /^・\w*/) {              # 例文
        if ($$inbuf =~ /(。|」)/) {         # 訳語あり
            FormSample(\@$outbuf, \$$inbuf);
#            PutLine(\@$outbuf, \$$inbuf);
        }
        else {
            if ($$inbuf =~ /",/) {          # 解説末尾の行
                PutLine(\@$outbuf, \$$inbuf);
            }
        }
    }
    else {
        PutLine(\@$outbuf, \$$inbuf);
    }
}

sub GetWord
{
    my $buff = $_[0];
    my $opt = $_[1];        # オプション：0以外で大文字を小文字に変換

    my $len = index($buff, '"', 1) - 1;
    my $entbuf = substr($buff, 1, $len);
    if ($opt) {
        $entbuf =~ tr/A-Z/a-z/;
    }
    $entbuf;
}

sub SelectPrint             # 同じ見出項目が２つある時、人名と地名を削除
{
    my ($orgbuf, $bakbuf) = @_;

    if ($$orgbuf[0] =~ /【(人|地)名/) {
        @$bakbuf = @$orgbuf;
    }
    else {                          # バッファ内の項目を書き出す
        if (@$bakbuf) {
            my $orgwd = GetWord($$orgbuf[0], 0);
            my $bakwd = GetWord($$bakbuf[0], 1);
            if ($orgwd eq $bakwd) {
#print "\t orgwd:{$orgwd}  bakwd:{$bakwd}\n";
#$dbg = <STDIN>;
                PrintBuff(\@$orgbuf);
            }
            else {
                PrintBuff(\@$bakbuf);
                PrintBuff(\@$orgbuf);
            }
            @$bakbuf = ();
        }
        else {
            PrintBuff(\@$orgbuf);
        }
    }
}

sub LinkLabel       # リンク項目の作成
{
    my ($buff) = @_;
    my $word = '';
    my $res = 0;

    if ($$buff =~ /(\w+)\sの過去/) {
        $word = $1;
        $$buff =~ s/\b$word\b/<→$&>/;
    }
    if ($$buff =~ /(\w+)\sの過去分詞/) {
        $word = $1;
        $$buff =~ s/\b$word\b/<→$&>/;
    }
    if ($$buff =~ /(\w+)\sの現在分詞/) {
        $word = $1;
        $$buff =~ s/\b$word\b/<→$&>/;
    }
    if ($$buff =~ /(\w+)\sの複数形/) {
        $word = $1;
        $$buff =~ s/\b$word\b/<→$&>/;
    }
    if ($word ne '') {
        $res = 1;
    }
    return $res;
}

sub PutInfl         # 変化=>原形データをバッファに詰める
{
    my ($list, $hash, $root, $word) = @_;

    push(@$list, $word);
    $$hash{$word} = $root;
    while (@$list > 30) {
        my $key = shift(@$list);
        delete $$hash{$key};
    }
}

sub GetInfl         # 変化=>原形リスト作成
{
    my ($list, $hash, $buff, $root) = @_;

    if ($buff =~ /【変化/) {
        if ($buff =~ /《動/) {
            if ($buff =~ /[\w\-]+ed/) {
                PutInfl(\@$list, \%$hash, $root, $&);
            }
            if ($buff =~ /[\w\-]+ing/) {
                PutInfl(\@$list, \%$hash, $root, $&);
            }
            if ($buff =~ /[\w\-]+s/) {
                PutInfl(\@$list, \%$hash, $root, $&);
            }
        }
        if ($buff =~ /《複/) {
            if ($buff =~ /[\w\-]+s/) {
                PutInfl(\@$list, \%$hash, $root, $&);
            }
        }
    }
}

sub GetGloss            # 見出の訳語をコピー
{
    my ($buff) = @_;
    my $gloss = '';

    if ($buff =~ /<→/ || $buff =~ /、/) {
        return $gloss;
    }
    if ($buff =~ /形】(.+?)の/) {
        $gloss = $1;
    }
    elsif ($buff =~ /形】(.+?)な/) {
        $gloss = $1;
    }
    elsif ($buff =~ /名】》(.+?)"/) {
        $gloss = $1;
    }
    elsif ($buff =~ /名】(.+?)"/) {
        $gloss = $1;
    }
    return $gloss;
}

sub PutGloss            # 見出語=>訳語をハッシュに入れる
{
    my ($hlist, $word, $gloss) = @_;

    %$hlist = ();
    $$hlist{$word} = $gloss;
}

sub SetGloss            # 削除項目用の見出語=>訳語ハッシュを作成
{
    my ($hlist, $buff) = @_;
    my $word = '';

    if ( $idiom_flg ) {
        return;
    }
    if ($buff =~ /[\w\-]+/) {
        $word = $&;
    }
    my $gloss = GetGloss($buff);
    if ($word && $gloss) {
        PutGloss(\%$hlist, $word, $gloss);
    }
}

while ($l = <IN>) {

    if (!($main_count % 2000)) {             # 処理表示カウンタ
        print STDERR "Line: $main_count\r";
    }
    $main_count++;

    if ($l =~ /^"/) {                        # 見出し単語のある行
        if ($l =~ /^",/ || $l =~ /^""/) {    # 説明中に改行の入った不正なパターン
            GetLine(\@out_buff, \$l);        # 読み込んだ項目をバッファにコピー
        }
        else {
            if (ChkItem(\@out_buff)) {
#               SelectPrint(\@out_buff, \@bak_buff);    # 未使用 （削除禁止）
                PrintBuff(\@out_buff);       # バッファ内の項目を書き出す
            }
            SetGloss(\%gloss_hash, $out_buff[0]);
            @out_buff = ();
            $link_flg = 0;
            $word_level = '0';

            if ($l =~ /[\w\-]+/) {
                $head_word = $&;
            }
            $idiom_flg = ChkIdiom($l);
            if ( !$link_flg ) {
                $link_flg = LinkLabel(\$l);
            }
            $word_level = GetLebel($l);
            if ( !$idiom_flg ) {            # 熟語でない単語&Levelを保存　'14.1.17
                if ($head_word !~ /\-/) {    # ハイフン入の単語は除く
                    $main_word = $head_word;
                    $main_level = $word_level;
                }
            }
            PutLine(\@out_buff, \$l);
        }
    }
    else {                                   # 見出しのない行
        if ( !$link_flg ) {
            $link_flg = LinkLabel(\$l);
        }
        if ( !$idiom_flg ) {
            GetInfl(\@infl_list, \%infl_hash, $l, $head_word);
        }
        if ($word_level eq '0') {
            $word_level = GetLebel($l);
        }

        if ( !$idiom_flg && $head_word !~ /\-/) {   # '14.1.17
            if ($main_word eq $head_word) {
                $main_level = $word_level;
            }
        }
        GetLine(\@out_buff, \$l);            # 読み込んだ項目をバッファにコピー
    }
}

print STDERR "Line: $main_count\r\n";

close(IN);
#close(OUT);

